from datetime import datetime

def datetime_to_timestamp(datetime_obj=datetime.now()):
    """
    If given a `datetime_obj`, converts it to milliseconds since epoch.
    Else, returns the milliseconds between now and the epoch.
    """
    epoch = datetime.utcfromtimestamp(0)
    delta = datetime_obj - epoch
    # Python 3-compatible TimeDelta.total_seconds()
    seconds = (delta.microseconds +
               (delta.seconds + delta.days * 24 * 3600) * 10 ** 6) / 10.0 ** 6
    milliseconds = seconds * 1000
    return int(milliseconds)



from collections import MutableMapping
import re

URL_PATTERNS = [
    "/v0/(?P<collection>.+)/(?P<key>.+)/events/(?P<type>.+)/(?P<timestamp>\d+)/(?P<ordinal>\d+)",
    "/v0/(?P<collection>.+)/(?P<key>.+)/events/(?P<type>.+)/(?P<timestamp>\d+)",
    "/v0/(?P<collection>.+)/(?P<key>.+)/events/(?P<type>.+)",
    "/v0/(?P<collection>.+)/(?P<key>.+)/refs/(?P<ref>.+)",
    "/v0/(?P<collection>.+)/(?P<key>.+)/refs",
    "/v0/(?P<collection>.+)/(?P<key>.+)/relations/(?P<kind>.+)/(?P<to_collection>.+)/(?P<to_key>.+)",
    "/v0/(?P<collection>.+)/(?P<key>.+)/relations/(?P<kinds>.+)",
    "/v0/(?P<collection>.+)/(?P<key>.+)",
    "/v0/(?P<collection>.+)"
]


class Response(MutableMapping):

    def __init__(self, resp):
        self.response = resp
        if self.response.text:
            self.json = self.response.json()
        else:
            self.json = dict()
        self._set_path()

    def _set_path(self):
        # match the url
        path = self.response.url[self.response.url.find('/v0'):]
        for regex in URL_PATTERNS:
            location_match = re.match(regex, path)
            if location_match:
                for key, value in location_match.groupdict().items():
                    setattr(self, key, value)
                break
        # match headers
        for regex in URL_PATTERNS:
            # check location
            location_match = re.match(
                regex, self.response.headers.get('location', ''))
            # if not in location, try content-location
            if not location_match:
                location_match = re.match(
                    regex, self.response.headers.get('content-location', ''))
            # if a match was found in either place, attach it to self
            if location_match:
                for key, value in location_match.groupdict().items():
                    setattr(self, key, value)
                break
        # finally, try the etag
        etag_match = re.match(
            '"(?P<ref>.+)"', self.response.headers.get('etag', ''))
        if etag_match:
            for key, value in etag_match.groupdict().items():
                setattr(self, key, value)

    def __getattr__(self, name):
        return getattr(self.response, name)

    def __getitem__(self, key):
        return self.json.get(key)

    def __setitem__(self, key, value):
        self.json[key] = value

    def __delitem__(self, key):
        del self.json[key]

    def __iter__(self):
        return iter(self.json)

    def __len__(self):
        return len(self.json)



from datetime import datetime
from .resource import Resource
from .version import VERSION
from .pages import Pages
from .patch import Patch
from .search import Search
from . import util

class Client(Resource):
    def __init__(self, api_key, url = None, use_async = False, **kwargs):
        self.api_key = api_key
        self.url = url

        # If no url is provided, use the default
        if url is None:
            self.url = 'https://api.orchestrate.io'

        if 'headers' not in kwargs:
            kwargs['headers'] = {
                'Content-Type': 'application/json',
                'Accept': 'application/json',
                'User-Agent': 'python-requests/1.2.0 porc/%s' % VERSION
            }
        kwargs['auth'] = (self.api_key, '')
        super(Client, self).__init__(self.url, use_async, **kwargs)

    def ping(self):
        return self._request('HEAD')

    def head(self, collection=None, key=None, ref=None):
        path = list()
        if collection:
            path.append(collection)
            if key:
                path.append(key)
                if ref:
                    path.append('refs')
                    path.append(ref)
        return self._request('HEAD', path)

    def get(self, collection, key, ref=None):
        if ref:
            path = [collection, key, 'refs', ref]
        else:
            path = [collection, key]
        return self._request('GET', path)

    def post(self, collection, body):
        return self._request('POST', [collection], body)

    def put(self, collection, key, body, ref=None):
        opts = dict()
        if ref:
            opts['If-Match'] = ref.center(len(ref) + 2, '"')
        elif ref == False:
            opts['If-None-Match'] = '"*"'
        return self._request('PUT', [collection, key], body, opts)

    def patch(self, collection, key, body_or_patch, ref=None):
        opts = {'Content-Type': 'application/json-patch+json'}
        if type(ref) == str:
            opts['If-Match'] = '"' + ref + '"'
        elif ref == False:
            # If-None-Match is not relevant for a PATCH request.
            opts['If-None-Match'] = '"*"'

        if isinstance(body_or_patch, Patch):
            body = body_or_patch.operations
        else:
            body = body_or_patch

        return self._request('PATCH', [collection, key], body, opts)

    def patch_merge(self, collection, key, body, ref=None):
        opts = {'Content-Type': 'application/merge-patch+json'}
        if type(ref) == str:
            opts['If-Match'] = '"' + ref + '"'
        elif ref == False:
            # If-None-Match is not relevant for a PATCH request.
            opts['If-None-Match'] = '"*"'

        return self._request('PATCH', [collection, key], body, opts)

    def delete(self, collection, key=None, ref=None):
        if key:
            opts = dict()
            params = dict()
            if ref:
                opts['If-Match'] = ref.center(len(ref) + 2, '"')
            else:
                params['purge'] = True
            return self._request('DELETE', [collection, key], params, opts)
        else:
            return self._request('DELETE', [collection], dict(force=True))

    def refs(self, collection, key, **params):
        return self._request('GET', [collection, key, 'refs'], params)

    def list(self, collection, **params):
        return Pages(self.opts, self.uri, [collection], params)

    def search(self, collection, query_or_search, **params):
        if isinstance(query_or_search, Search):
            params.update(query_or_search.params)
        else:
            params['query'] = query_or_search

        return Pages(self.opts, self.uri, [collection], params)

    def get_relations(self, collection, key, *relations):
        path = [collection, key, 'relations'] + list(relations)
        return self._request('GET', path)

    def put_relation(self, collection, key, relation, to_collection, to_key):
        path = [collection, key, 'relation', relation, to_collection, to_key]
        return self._request('PUT', path)

    def delete_relation(self, collection, key, relation, to_collection, to_key):
        path = [collection, key, 'relation', relation, to_collection, to_key]
        return self._request('DELETE', path, dict(purge=True))

    def get_event(self, collection, key, event_type, timestamp, ordinal):
        if isinstance(timestamp, datetime):
            timestamp = util.datetime_to_timestamp(timestamp)
        path = [collection, key, 'events', event_type, timestamp, ordinal]
        return self._request('GET', path)

    def post_event(self, collection, key, event_type, data, timestamp=None):
        path = [collection, key, 'events', event_type]
        if timestamp:
            if isinstance(timestamp, datetime):
                timestamp = util.datetime_to_timestamp(timestamp)
            path.append(timestamp)
        return self._request('POST', path, data)

    def put_event(self, collection, key, event_type, timestamp, ordinal, data, ref=None):
        if isinstance(timestamp, datetime):
            timestamp = util.datetime_to_timestamp(timestamp)
        path = [collection, key, 'events', event_type, timestamp, ordinal]
        headers = dict()
        if ref:
            headers['If-Match'] = ref.center(len(ref) + 2, '"')
        return self._request('PUT', path, data, headers=headers)

    def delete_event(self, collection, key, event_type, timestamp, ordinal, ref=None):
        if isinstance(timestamp, datetime):
            timestamp = util.datetime_to_timestamp(timestamp)
        path = [collection, key, 'events', event_type, timestamp, ordinal]
        headers = dict()
        params = dict(purge=True)
        if ref:
            headers['If-Match'] = ref.center(len(ref) + 2, '"')
        return self._request('DELETE', path, params, headers=headers)

    def list_events(self, collection, key, event_type, **params):
        path = [collection, key, 'events', event_type]
        for param in ['startEvent', 'afterEvent', 'beforeEvent', 'endEvent']:
            if param in params and isinstance(params[param], datetime):
                params[param] = util.datetime_to_timestamp(params[param])
        return Pages(self.opts, self.uri, path, params)

    def async(self):
        return Async(self.api_key, self.url, **self.opts)


class Async(Client):

    def __init__(self, api_key, url, **opts):
        super(Async, self).__init__(api_key, url, True, **opts)

    def __enter__(self):
        return self

    def __exit__(self, type, value, stacktrace):
        return True



class Search:
    def __init__(self):
        self.params = {}

    def limit(self, limit):
        self.params['limit'] = limit
        return self

    def offset(self, offset):
        self.params['offset'] = offset
        return self

    def aggregate(self, aggregate_type, field, options=None):
        aggregate = "value.%s:%s" % (field, aggregate_type)

        if type(options) is list:
            aggregate += ":%s" % ':'.join(options)
        elif type(options) is str:
            aggregate += ":%s" % options

        if 'aggregate' in self.params:
            self.params['aggregate'] = ','.join([self.params['aggregate'], aggregate])
        else:
            self.params['aggregate'] = aggregate

        return self

    def sort(self, field, order):
        sort = "value.%s:%s" % (field, order)

        if 'sort' in self.params:
            self.params['sort'] = ','.join([self.params['sort'], sort])
        else:
            self.params['sort'] = sort

        return self

    def query(self, query):
        self.params['query'] = query
        return self



from .resource import Resource
from collections import Iterator

class Pages(Iterator):
    def __init__(self, opts, url, path, params):
        self.initialPath = path
        self.nextPath = path
        self.prevPath = None
        self.resource = Resource(url, **opts)
        self.params = params

    def _move(self, path, querydict = {}, **headers):
        if path is None:
            raise StopIteration

        # Make a copy of parameters
        params = self.params.copy()
        params.update(querydict)

        # Get the page
        response = self.resource._request('GET', path, params, **headers)

        # Extract the next/prev links
        self.nextPath = response.links.get('next', {}).get('url')
        self.prevPath = response.links.get('prev', {}).get('url')
        return response

    def reset(self):
        """
        Clear the page's current place.

            page_1 = page.next().result()
            page_2 = page.next().result()
            page.reset()
            page_x = page.next().result()
            assert page_x.url == page_1.url
        """
        self.nextPath = self.initialPath
        self.prevPath = None

    def next(self, querydict={}, **headers):
        """
        Gets the next page of results.
        Raises `StopIteration` when there are no more results.
        """
        return self._move(self.nextPath, querydict, **headers)

    def __next__(self):
        return self.next()

    def prev(self, querydict={}, **headers):
        """
        Gets the previous page of results.
        Raises `StopIteration` when there are no more results.

        Note: Only collection searches provide a `prev` value.
        For all others, `prev` will always return `StopIteration`.
        """
        return self._move(self.prevPath, querydict, **headers)

    def all(self):
        results = []
        for response in self:
            response.raise_for_status()
            results.extend(response['results'])
        return results



import json
import requests
from .response import Response
from requests_futures.sessions import FuturesSession

try:
    # python 2
    from urllib import quote
    from urlparse import urljoin
except ImportError:
    # python 3
    from urllib.parse import quote, urljoin


class Resource(object):
    def __init__(self, uri, use_async=False, **kwargs):
        self.uri = uri
        self.opts = kwargs
        self.session = requests.Session()
        self.async_session = FuturesSession()
        self.use_async = use_async
        kwargs['hooks'] = {
            "response": self._handle_response
        }
        for obj in [self.session, self.async_session]:
            for key, value in kwargs.items():
                setattr(obj, key, value)

    def _make_path(self, path):
        # If the first element of the list isn't v0, insert it
        if len(path) > 0 and path[0] != "v0":
            path.insert(0, "v0")
        elif len(path) == 0:
            path = ["v0"]

        # Escape the components of the path
        return '/'.join([quote(str(elem), '') for elem in path])

    def _request(self, method, path = [], body = None, headers = {}):
        """
        Executes the request based on the given body and headers
        along with options set on the object.
        """
        if isinstance(path, list):
            path = self._make_path(path)

        uri = urljoin(self.uri, path)

        opts = dict(headers=headers)
        session = self.async_session if self.use_async else self.session
        # normalize body according to method and type
        if body != None:
            if method.lower() in ['head', 'get', 'delete']:
                if type(body) == dict:
                    # convert True and False to true and false
                    for key, value in list(body.items()):
                        if value is True:
                            body[key] = 'true'
                        elif value is False:
                            body[key] = 'false'
                opts['params'] = body
            else:
                opts['data'] = json.dumps(body)

        return session.request(method, uri, **opts)

    def _handle_response(self, response, *args, **kwargs):
        return Response(response)
